require 'uu/os/http/http_request'
require 'uu/os/http/http_headers'

module UU
  module OS
    module Server

      # Ancestor of all servlets (low-level components allowing to process
      # incoming request and provide custom response without direct dependency
      # on particular server implementation).
      class AbstractServlet

        # Main method for processing requests. Response should be built using
        # methods {#response}, {#file_download_response} or {#streamed_response}.
        # @param [UU::OS::HTTP::HTTPRequest] request Incoming request.
        # @return [Object] Response provided by {#response}, {#file_download_response}
        #    or {#streamed_response} methods.
        # @note Method must be implemented in each servlet inherited from this class!
        # @private
        def service(request) # :nodoc:
          raise NotImplementedError, "method not implemented for #{self.class}"
        end

        # Method for construction of custom response.
        # @param [Fixnum] status HTTP status code (see {http://en.wikipedia.org/wiki/List_of_HTTP_status_codes
        #    List of HTTP status codes} for details).
        # @param [Hash, UU::OS::HTTP::HTTPHeaders] headers Response headers (multi-value headers should
        #    be given as array value).
        # @param [String, IO] body Response body.
        # @return [Object] Response object to be handled by server.
        def response(status, headers, body)
          if !headers.nil? && !headers.kind_of?(Hash)
            raise "invalid headers, expected hash but was #{headers.class}"
          end
          if !body.nil? && !body.kind_of?(String) && !body.respond_to?(:read)
            raise "cannot send #{body.class} as response"
          end
          headers ||= {}
          headers = headers.inject({}) do |tmp, (key, value)|
            key = key.to_s.gsub('_', '-').downcase
            if value.kind_of?(Array)
              tmp[key] = value.join(',')
            else
              tmp[key] = value.to_s
            end
            tmp
          end
          headers['content-type'] ||= 'application/octet-stream'
          return {status: status.to_i, headers: headers, body: body}
        end

        # Constructs file download response with HTTP status code 200
        # (should be used to return binary data).
        # @param [String, IO] body Object representing binary content.
        # @param [Hash] meta_data Additional optional response attributes
        #    (if not set, they are obtained from body if body provides them):
        # @option meta_data [String] :name File name
        # @option meta_data [String] :size Response content size
        # @option meta_data [String] :content_type Content type
        # @param [Proc] block Optional block with IO parameter for asynchronous
        #    streaming of binary content (if given, +body+ attribute is ignored).
        # @return [Object] Response object to be handled by server.
        def file_download_response(body = nil, meta_data = nil, &block)
          body, meta_data = meta_data, body if body.kind_of?(Hash)
          if !body.nil? && !body.kind_of?(String) && !body.respond_to?(:read)
            raise "cannot send #{body.class} as response"
          end
          meta_data ||= {}
          meta_data[:name] ||= File.basename(body) if body.respond_to?(:path)
          meta_data[:size] ||= body.size if body.respond_to?(:size)
          meta_data[:content_type] ||= body.content_type if body.respond_to?(:content_type)
          headers = {}
          headers[:content_type] = meta_data[:content_type] || 'application/octet-stream'
          headers[:content_disposition] = 'attachment'
          headers[:content_disposition] << "; filename=\"#{meta_data[:name]}\"" if meta_data[:name]
          headers[:content_length] = meta_data[:size] if meta_data[:size]
          if block.nil?
            return response(200, headers, body)
          else
            return streamed_response(headers, &block)
          end
        end

        # Method allows asynchronous streaming of body with HTTP status code 200.
        # @param [Hash, UU::OS::HTTP::HTTPHeaders] headers Response headers
        #    (multi-value headers should be given as array value).
        # @param [Proc] block Block with IO parameter for asynchronous streaming.
        # @return [Object] Response object to be handled by server.
        def streamed_response(headers, &block)
          raise 'no block' if block.nil?
          read, write = IO.pipe
          write.write('')
          Thread.new {
            block.yield(write)
          }
          return response(200, headers, read)
        end

      end

    end
  end
end
